C$Procedure DSKHAS ( DSK, does DSK segment have attribute? )
 
      LOGICAL FUNCTION DSKHAS ( FLAG, FLGIDX, ATTFLG )
  
C$ Abstract
C
C     Indicate whether a given DSK segment descriptor attribute flag is
C     set.
C
C$ Disclaimer
C
C     THIS SOFTWARE AND ANY RELATED MATERIALS WERE CREATED BY THE
C     CALIFORNIA INSTITUTE OF TECHNOLOGY (CALTECH) UNDER A U.S.
C     GOVERNMENT CONTRACT WITH THE NATIONAL AERONAUTICS AND SPACE
C     ADMINISTRATION (NASA). THE SOFTWARE IS TECHNOLOGY AND SOFTWARE
C     PUBLICLY AVAILABLE UNDER U.S. EXPORT LAWS AND IS PROVIDED "AS-IS"
C     TO THE RECIPIENT WITHOUT WARRANTY OF ANY KIND, INCLUDING ANY
C     WARRANTIES OF PERFORMANCE OR MERCHANTABILITY OR FITNESS FOR A
C     PARTICULAR USE OR PURPOSE (AS SET FORTH IN UNITED STATES UCC
C     SECTIONS 2312-2313) OR FOR ANY PURPOSE WHATSOEVER, FOR THE
C     SOFTWARE AND RELATED MATERIALS, HOWEVER USED.
C
C     IN NO EVENT SHALL CALTECH, ITS JET PROPULSION LABORATORY, OR NASA
C     BE LIABLE FOR ANY DAMAGES AND/OR COSTS, INCLUDING, BUT NOT
C     LIMITED TO, INCIDENTAL OR CONSEQUENTIAL DAMAGES OF ANY KIND,
C     INCLUDING ECONOMIC DAMAGE OR INJURY TO PROPERTY AND LOST PROFITS,
C     REGARDLESS OF WHETHER CALTECH, JPL, OR NASA BE ADVISED, HAVE
C     REASON TO KNOW, OR, IN FACT, SHALL KNOW OF THE POSSIBILITY.
C
C     RECIPIENT BEARS ALL RISK RELATING TO QUALITY AND PERFORMANCE OF
C     THE SOFTWARE AND ANY RELATED MATERIALS, AND AGREES TO INDEMNIFY
C     CALTECH AND NASA FOR ALL THIRD-PARTY CLAIMS RESULTING FROM THE
C     ACTIONS OF RECIPIENT IN THE USE OF THE SOFTWARE.
C
C$ Required_Reading
C
C     DAS
C     DSK
C
C$ Keywords
C
C     DAS
C     DSK
C     FILES
C
C$ Declarations

      IMPLICIT NONE
      INCLUDE 'dskdsc.inc'

      INTEGER               FLAG
      INTEGER               FLGIDX
      INTEGER               ATTFLG

C$ Brief_I/O
C
C     Variable  I/O  Description
C     --------  ---  --------------------------------------------------
C     FLAG       I   A flag element of a DSK descriptor.
C     FLGIDX     I   Index of FLAG in array of flag elements.
C     ATTFLG     I   Index of attribute bit within FLAG.
C
C     The function returns a logical flag indicating whether the 
C     specified attribute bit is set.
C
C$ Detailed_Input
C
C     FLAG           is a flag element of a DSK descriptor.
C                    DSK descriptors contain multiple integers
C                    that act as bit fields representing 
C                    collections of logical flags.
C
C     FLGIDX         is the index of FLAG in the array of
C                    flag elements of a DSK descriptor.
C
C     ATTFLG         is the index of a bit within FLAG that
C                    represents that presence of an attribute
C                    of interest.
C
C
C$ Detailed_Output
C
C     The function returns a logical flag indicating whether the 
C     specified attribute bit is set.
C
C$ Parameters
C
C     See the INCLUDE file
C
C         dskdsc.inc
C
C$ Exceptions
C
C     1) If FLGIDX doesn't refer to a valid flag
C        element of a DSK descriptor, the error 
C        SPICE(VALUEOUTOFRANGE) is signaled.
C
C     2) If FLGIDX doesn't refer to a currently used flag
C        element of a DSK descriptor, the error 
C        SPICE(NOTSUPPORTED) is signaled.
C
C     3) If ATTFLG doesn't refer to a valid bit
C        index within FLAG, the error
C        SPICE(NOTSUPPORTED) is signaled.
C
C$ Files
C
C     See input argument HANDLE.
C
C$ Particulars
C
C     This is a utility that encapsulates logic required
C     to interpret DSK descriptor flag elements.
C
C$ Examples
C
C     1) Determine whether a DSK segment contains normal
C        vectors.
C
C            INCLUDE 'dskdsc.inc'
C
C            DOUBLE PRECISION      DSKDSC ( DSKDSZ )
C            LOGICAL               DSKHAS
C            LOGICAL               STATUS
C
C                [ Obtain value of DSKDSC ]
C
C            STATUS = DSKHAS( INT(DSKDSC(FL1IDX)), 1, FL1NRM )
C            WRITE (*,*) 'Normals included: ', STATUS
C
C            END
C
C
C$ Restrictions
C
C     1) This is a prototype routine. 
C
C$ Literature_References
C
C     None.
C
C$ Author_and_Institution
C
C     N.J. Bachman    (JPL)
C
C$ Version
C
C-    SPICELIB Version 1.0.0, 09-OCT-2009 (NJB)
C
C-&
 
C$ Index_Entries
C
C     fetch attribute flags from a type 2 dsk segment
C
C-&
 

C
C     SPICELIB functions
C
      LOGICAL               ODD
      LOGICAL               RETURN

C
C     Local variables
C

      
C
C     Start out assuming attribute is not present.
C
      DSKHAS = .FALSE.

      IF ( RETURN() ) THEN         
         RETURN
      END IF

      CALL CHKIN ( 'DSKHAS' )

      IF (  ( FLGIDX .LT. 0 ) .OR. ( FLGIDX .GT. 2 )  ) THEN

         CALL SETMSG ( 'DSK descriptors contain only two flag sets; ' //
     .                 'flag index was #.'                            )
         CALL ERRINT ( '#',  FLGIDX                                   )
         CALL SIGERR ( 'SPICE(VALUEOUTOFRANGE)'                       )
         CALL CHKOUT ( 'DSKHAS'                                       )
         RETURN      

      END IF

      
      IF ( FLGIDX .EQ. 1 ) THEN

         IF ( ATTFLG .EQ. FL13DT ) THEN
C
C           Set the return value based on the 3-D topography bit.
C
            DSKHAS = ODD ( FLAG / FL13DT )


         ELSE IF ( ATTFLG .EQ. FL1NRM ) THEN
C
C           Set the return value based on the normal vector bit.
C
            DSKHAS = ODD ( FLAG / FL1NRM )

            WRITE (*,*) FLAG, FL1NRM, FLAG/FL1NRM

         ELSE IF ( ATTFLG .EQ. FL1ALB ) THEN
C
C           Set the return value based on the albedo bit.
C
            DSKHAS = ODD ( FLAG / FL1ALB )

         ELSE IF ( ATTFLG .EQ. FL1EQP ) THEN
C
C           Set the return value based on the equipotential bit.
C
            DSKHAS = ODD ( FLAG / FL1EQP )
        
         ELSE

            CALL SETMSG ( 'Unrecognized attribute flag: # ' )
            CALL ERRINT ( '#',  ATTFLG                      )
            CALL SIGERR ( 'SPICE(NOTSUPPORTED)'             )
            CALL CHKOUT ( 'DSKHAS'                          )
            RETURN

         END IF

      ELSE

         CALL SETMSG ( 'Currently only the first flag is used, ' //
     .                 'but flag index was #.'                   )
         CALL ERRINT ( '#',  FLGIDX                              )
         CALL SIGERR ( 'SPICE(NOTSUPPORTED)'                     )
         CALL CHKOUT ( 'DSKHAS'                                  )
         RETURN

      END IF

C
C     Currently the second flag set is unused.
C
      CALL CHKOUT ( 'DSKHAS' )
      RETURN      
      END
