C$Procedure DSKX02 ( DSK, ray-surface intercept, type 2 )

      SUBROUTINE DSKX02 ( HANDLE, DLADSC, VERTEX, 
     .                    RAYDIR, PLID,   XPT,    FOUND )
 
C$ Abstract
C
C     Determine the plate ID and body-fixed coordinates of the
C     intersection of a specified ray with the surface defined by a
C     type 2 DSK plate model.
C
C$ Disclaimer
C
C     THIS SOFTWARE AND ANY RELATED MATERIALS WERE CREATED BY THE
C     CALIFORNIA INSTITUTE OF TECHNOLOGY (CALTECH) UNDER A U.S.
C     GOVERNMENT CONTRACT WITH THE NATIONAL AERONAUTICS AND SPACE
C     ADMINISTRATION (NASA). THE SOFTWARE IS TECHNOLOGY AND SOFTWARE
C     PUBLICLY AVAILABLE UNDER U.S. EXPORT LAWS AND IS PROVIDED "AS-IS"
C     TO THE RECIPIENT WITHOUT WARRANTY OF ANY KIND, INCLUDING ANY
C     WARRANTIES OF PERFORMANCE OR MERCHANTABILITY OR FITNESS FOR A
C     PARTICULAR USE OR PURPOSE (AS SET FORTH IN UNITED STATES UCC
C     SECTIONS 2312-2313) OR FOR ANY PURPOSE WHATSOEVER, FOR THE
C     SOFTWARE AND RELATED MATERIALS, HOWEVER USED.
C
C     IN NO EVENT SHALL CALTECH, ITS JET PROPULSION LABORATORY, OR NASA
C     BE LIABLE FOR ANY DAMAGES AND/OR COSTS, INCLUDING, BUT NOT
C     LIMITED TO, INCIDENTAL OR CONSEQUENTIAL DAMAGES OF ANY KIND,
C     INCLUDING ECONOMIC DAMAGE OR INJURY TO PROPERTY AND LOST PROFITS,
C     REGARDLESS OF WHETHER CALTECH, JPL, OR NASA BE ADVISED, HAVE
C     REASON TO KNOW, OR, IN FACT, SHALL KNOW OF THE POSSIBILITY.
C
C     RECIPIENT BEARS ALL RISK RELATING TO QUALITY AND PERFORMANCE OF
C     THE SOFTWARE AND ANY RELATED MATERIALS, AND AGREES TO INDEMNIFY
C     CALTECH AND NASA FOR ALL THIRD-PARTY CLAIMS RESULTING FROM THE
C     ACTIONS OF RECIPIENT IN THE USE OF THE SOFTWARE.
C
C$ Required_Reading
C
C      None.
C
C$ Keywords
C
C      GEOMETRY
C      SHAPES
C
C$ Declarations
 
      IMPLICIT NONE
 
      INCLUDE              'pltmax.inc'
      INCLUDE              'dla.inc'
      INCLUDE              'dskdsc.inc'
      INCLUDE              'dsk02.inc'
 
      INTEGER               HANDLE
      INTEGER               DLADSC   ( * )
      DOUBLE PRECISION      VERTEX   ( 3 )
      DOUBLE PRECISION      RAYDIR   ( 3 )
      INTEGER               PLID
      DOUBLE PRECISION      XPT      ( 3 )
      LOGICAL               FOUND
 
C$ Brief_I/O
C
C     VARIABLE  I/O  DESCRIPTION
C     --------  ---  --------------------------------------------------
C     HANDLE     I   Handle of DSK kernel containing plate model.
C     DLADSC     I   DLA descriptor of plate model segment.
C     VERTEX     I   Ray's vertex in the  body fixed frame.
C     RAYDIR     I   Ray direction in the body fixed frame.
C     PLID       O   ID code of the plate intersected by the ray.
C     XPT        O   Intercept.
C     FOUND      O   Flag indicating whether intercept exists.
C
C$ Detailed_Input
C
C     HANDLE     is the file handle of a DSK file containing a shape
C                model for a target body.  The shape model is stored
C                in a type 2 DSK segment.
C   
C     DLADSC     is the DLA descriptor of a type 2 DSK segment
C                containing plate model data representing the surface of
C                the target body.  The caller should declare DLADSC
C                with size DLADSZ; this size parameter is defined in
C                the INCLUDE file dla.inc.  Normally this descriptor
C                will be obtained by a search through a DSK file
C                using the DLA search routines; see the Examples
C                header section below for a working code example
C                illustrating a simple search.
C
C     VERTEX     is the vertex of a ray.  VERTEX is expressed relative
C                to the body fixed reference frame associated with the
C                target body.  This reference frame is the same frame
C                relative to which the vertices of the plate model
C                are expressed.  Units are km.
C
C                The vertex is not required to be outside the target
C                body.
C
C     RAYDIR     is the ray's direction vector.  RAYDIR is expressed
C                relative to the body fixed reference frame associated
C                with the target body.
C
C$ Detailed_Output
C
C     PLID       is the ID of the plate closest to the input ray's
C                vertex at which a ray-surface intercept exists.
C                If no intercept exists, PLID is undefined.
C
C     XPT        is the ray-target intercept closest to the ray's
C                vertex, if an intercept exists.  XPT is expressed
C                relative to the body-fixed reference frame associated
C                with the target body.  Units are km.
C
C                If no intercept exists, XPT is undefined.
C
C     FOUND      is a logical flag that indicates whether or not the
C                ray does indeed intersect the target.  If the ray
C                intersects a plate FOUND is .TRUE.  Otherwise FOUND is
C                FALSE.
C
C$ Parameters
C
C     See INCLUDE files
C
C        pltmax.inc
C        dla.inc
C        dskdsc.inc
C        dsk02.inc
C
C$ Exceptions
C
C     1)  If the input handle is invalid, the error will be diagnosed by
C         routines in the call tree of this routine.
C
C     2)  If a file read error occurs, the error will be diagnosed by
C         routines in the call tree of this routine.
C
C     3)  If the input DLA descriptor is invalid, the effect of this
C         routine is undefined. The error *may* be diagnosed by routines
C         in the call  tree of this routine, but there are no
C         guarantees.
C
C     4)  If an error occurs while trying to look up any component
C         of the shape model, the error will be diagnosed by routines
C         in the call tree of this routine.
C
C     5)  If the input ray direction is the zero vector, the error
C         SPICE(ZEROVECTOR) is signaled.
C
C     6)  If the coarse voxel grid scale of the shape model is less than
C         1, the error SPICE(VALUEOUTOFRANGE) is signaled.
C    
C     7)  If the coarse voxel grid of the shape model contains more
C         than MAXCGR (see pltmax.inc) voxels, the error
C         SPICE(GRIDTOOLARGE) is signaled.
C
C     8)  If the plate list for any intersected voxel is too large
C         for this routine to buffer, the error SPICE(ARRAYTOOSMALL)
C         is signaled.
C
C     9)  Due to round-off errors, results from this routine may
C         differ across platforms.  Results also may differ from
C         those expected---and not necessarily by a small amount.
C         For example, a ray may miss a plate it was expected to
C         hit and instead hit another plate considerably farther
C         from the ray's vertex, or miss the target entirely.
C
C     10) In the event that an intercept point lies on multiple 
C         plates (that is, the point is on an edge or vertex), 
C         a plate will be selected.  Due to round-off error, the
C         selection may vary across platforms.
C
C$ Files
C
C     See the description of the input argument HANDLE.
C
C$ Particulars
C    
C     This routine solves the ray-surface intercept problem for 
C     a specified ray and a surface represented by triangular plate 
C     model.  The surface representation is provided by data in a
C     type 2 segment of a DSK file.
C
C     This routine does not assume that the segment from which the
C     surface model data are read represents the entire surface of
C     the target body.  A program could call this routine repeatedly
C     to find the surface intercept of a ray and a shape model
C     partitioned into multiple segments.
C
C     In general, this routine should be expected to run faster
C     when used with smaller shape models.  
C
C$ Examples
C
C     1) Find the surface intercept points corresponding to a latitude/
C        longitude grid of a specified resolution, for a specified
C        target body.  This simple program assumes the shape model for
C        the target body is stored in a single type 2 DSK segment, and
C        that this segment is the first one in the DSK file to which it
C        belongs.
C
C                 PROGRAM LLGRID
C                 IMPLICIT NONE
C
C                 INCLUDE 'dla.inc'
C                 INCLUDE 'dskdsc.inc'
C                 INCLUDE 'dsk02.inc'
C
C           C
C           C     SPICELIB functions
C           C
C                 DOUBLE PRECISION      PI
C                 DOUBLE PRECISION      RPD
C
C           C
C           C     Local parameters
C           C 
C                 INTEGER               FILSIZ
C                 PARAMETER           ( FILSIZ = 255 )
C
C                 INTEGER               NLAT
C                 PARAMETER           ( NLAT   = 9 )
C
C                 INTEGER               NLON
C                 PARAMETER           ( NLON   = 9 )
C
C           C
C           C     Local parameters
C           C 
C                 DOUBLE PRECISION      TOL
C                 PARAMETER           ( TOL   =  1.D-12 )
C
C           C
C           C     Local variables
C           C 
C                 CHARACTER*(FILSIZ)    DSK
C
C                 DOUBLE PRECISION      LAT
C                 DOUBLE PRECISION      LON
C                 DOUBLE PRECISION      MAXR
C                 DOUBLE PRECISION      R
C                 DOUBLE PRECISION      RAYDIR ( 3 )
C                 DOUBLE PRECISION      VERTEX ( 3 )
C                 DOUBLE PRECISION      XLAT
C                 DOUBLE PRECISION      XLON
C                 DOUBLE PRECISION      XPT    ( 3 )
C                 DOUBLE PRECISION      XR
C
C                 INTEGER               DLADSC ( DLADSZ )
C                 INTEGER               HANDLE
C                 INTEGER               I
C                 INTEGER               J
C                 INTEGER               N
C                 INTEGER               PLID
C
C                 LOGICAL               FOUND
C
C           C
C           C     Prompt for the name of the DSK to read.
C           C
C                 CALL PROMPT ( 'Enter DSK name > ', DSK )
C           C
C           C     Open the DSK file for read access.
C           C     We use the DAS-level interface for
C           C     this function.
C           C
C                 CALL DASOPR ( DSK, HANDLE )
C           C
C           C     Begin a forward search through the
C           C     kernel, treating the file as a DLA.
C           C     In this example, it's a very short
C           C     search.
C           C
C                 CALL DLABFS ( HANDLE, DLADSC, FOUND )
C
C                 IF ( .NOT. FOUND ) THEN
C           C
C           C        We arrive here only if the kernel
C           C        contains no segments.  This is 
C           C        unexpected, but we're prepared for it.
C           C
C                    CALL SETMSG ( 'No segments found '
C                .   //            'in DSK file #.'    )
C                    CALL ERRCH  ( '#',  DSK           )
C                    CALL SIGERR ( 'SPICE(NODATA)'     )
C
C                 END IF
C
C           C
C           C     If we made it this far, DLADSC is the
C           C     DLA descriptor of the first segment.
C           C
C           C     We're going to generate the intercept points
C           C     using a set of rays which point toward the
C           C     origin and whose vertices are on a specified
C           C     specified lat/lon grid.  To start out we
C           C     must pick a reasonable range from the origin
C           C     for the vertices:  the range must be large 
C           C     enough so that the vertices are guaranteed
C           C     to be outside the target body but small 
C           C     enough that we don't lose too much precision
C           C     in the surface intercept computation.
C           C
C           C     We'll look up the upper bound for the target
C           C     radius, then use 2 times this value as the
C           C     vertex magnitude.
C           C
C                 CALL DSKD02 ( HANDLE, DLADSC, KWMAXR, 1, 1, N, MAXR )
C            
C                 R = 2.D0 * MAXR
C
C           C
C           C     Now generate the intercept points.  We generate
C           C     intercepts along latitude bounds, working from 
C           C     north to south.  The latitude range is selected
C           C     to range from +80 to -80 degrees.  Longitude
C           C     ranges from 0 to 320 degrees.  The increment
C           C     is 20 degrees for latitude and 40 degrees for
C           C     longitude.
C           C
C                 DO I = 1, NLAT
C
C                    LAT = RPD() * ( 100.D0 - 20.D0*I )
C
C                    DO J = 1, NLON
C                        
C                       LON = RPD() * 40.D0 * (J-1)
C           C
C           C           Produce a ray vertex for the current
C           C           lat/lon value.  Negate the vertex to
C           C           produce the ray's direction vector.
C           C
C                       CALL LATREC ( R, LON, LAT, VERTEX )
C                       CALL VMINUS ( VERTEX,      RAYDIR )
C           C
C           C           Find the surface intercept for this
C           C           ray. 
C           C
C                       CALL DSKX02 ( HANDLE, DLADSC, VERTEX,
C                .                    RAYDIR, PLID,   XPT,    FOUND  ) 
C           C
C           C           Since the ray passes through the origin on
C           C           the body-fixed frame associated with the 
C           C           target body, we'd rarely expect to find that
C           C           the ray failed to intersect the surface.
C           C           For safety, we check the FOUND flag.  (A
C           C           "not found" condition could be a sign of
C           C           a bug.)
C           C
C                       IF ( .NOT. FOUND ) THEN
C
C                          WRITE(*,*) ' '
C                          WRITE(*,*) 'Intercept not found!'
C                          WRITE(*,*) '   Ray vertex:'
C                          WRITE(*,*) '   Longitude (deg): ', LON/RPD()
C                          WRITE(*,*) '   Latitude  (deg): ', LAT/RPD()
C                          WRITE(*,*) '   Range      (km): ', R
C                          WRITE(*,*) ' '
C
C                       ELSE
C           C
C           C              This is the normal case.  Display the
C           C              intercept plate ID and the intercept
C           C              point in both cartesian and latitudinal
C           C              coordinates.  Show the corresponding ray
C           C              vertex to facilitate validation of results.
C           C              
C           C              Use RECRAD rather than RECLAT to produce
C           C              non-negative longitudes.
C           C
C                          CALL RECRAD ( XPT, XR, XLON, XLAT )
C
C                          WRITE(*,*) ' '
C                          WRITE(*,*) 'Intercept found:'
C                          WRITE(*,*) '   Plate ID:             ', PLID
C                          WRITE(*,*) '   Cartesian coordinates:', XPT
C                          WRITE(*,*) '   Latitudinal coordinates:'
C                          WRITE(*,*) '   Longitude (deg): ', XLON/RPD()
C                          WRITE(*,*) '   Latitude  (deg): ', XLAT/RPD()
C                          WRITE(*,*) '   Range      (km): ', XR
C                          WRITE(*,*) 
C                          WRITE(*,*) '   Ray vertex:'
C                          WRITE(*,*) '   Longitude (deg): ', LON/RPD()
C                          WRITE(*,*) '   Latitude  (deg): ', LAT/RPD()
C                          WRITE(*,*) '   Range      (km): ', R
C                          WRITE(*,*) ' '                       
C           C
C           C              Perform sanity checks on the intercept
C           C              coordinates.  Stop the program if any error
C           C              is larger than our tolerance value.
C           C                          
C                          IF ( ABS(XLAT-LAT) .GT. TOL ) THEN
C                             CALL SIGERR ( 'Latitude error!' )
C                          END IF   
C
C                          IF ( XLON-LON .GT. PI() ) THEN
C                             XLON = XLON-2*PI()
C                          END IF
C
C                          IF ( ABS(XLON-LON) .GT. TOL ) THEN
C                             CALL SIGERR ( 'Longitude error!' )
C                          END IF   
C
C                          IF ( XR .GE. (1+TOL)*MAXR ) THEN
C                             CALL SIGERR ( 'Range error!' )
C                          END IF
C
C                       END IF
C
C                    END DO
C
C                 END DO
C           C
C           C     Close the kernel.  This isn't necessary in a stand-
C           C     alone program, but it's good practice in subroutines
C           C     because it frees program and system resources.
C           C
C                 CALL DASCLS ( HANDLE )
C                 END
C
C
C$ Restrictions
C
C     1) This is prototype code.  When this code is graduated into
C        SPICELIB, the functionality or interface could change.
C
C$ Literature_References
C
C     Woo, A. "Fast Ray-Box Intersection", Graphic Gems I, 395-396.
C
C$ Author_and_Institution
C
C     N.J. Bachman    (JPL)
C     J.A. Bytof      (JPL)
C     W.L. Taber      (JPL)
C     E.D. Wright     (JPL)
C
C$ Version
C
C-    SPICELIB Version 1.0.1, 08-OCT-2009 (NJB)
C
C        Updated header.
C
C-    PLATELIB Version 2.0.0 10-FEB-2007 (NJB) 
C
C        Added SAVE statements to work around memory problems under 
C        Solaris/gcc.  
C
C-    PLATELIB Version 2.0.0 31-OCT-2006 (NJB) (EDW)
C
C        Modified to work with DLA-based kernels.  Many
C        changes were made to the algorithm to improve
C        execution speed.
C
C-    PLATELIB Version 1.1.0, 19-AUG-2004 (EDW)
C
C        Implemented "Fast Ray-Box Intersection" algorithm.
C        Renamed routine DSKX02 from PLBORE_3.
C
C-    PLATELIB Version 1.0.0, 25-FEB-1999 (JAB)
C
C        Based on PLBORE and PLBORE_2.
C
C-&
 
C$ Index_Entries
C
C     plate and plate model point intersected by ray
C
C-&

C$ Revisions
C
C-    PLATELIB Version 2.0.0 10-FEB-2007 (NJB) 
C
C        Added SAVE statements to work around memory problems under 
C        Solaris/gcc.  Variables that are now saved (although the
C        routine's logic does not require them to be) are:
C
C           ORDVEC
C           PLATID
C           SOURCE
C           VOXLST
C           VXLCG
C           VXLOUT
C           VXLSTR
C
C-&


C
C     SPICELIB Functions
C 
      DOUBLE PRECISION      DPMAX
      DOUBLE PRECISION      TOUCHD
      DOUBLE PRECISION      VDOT
      DOUBLE PRECISION      VNORM
          

      LOGICAL               RETURN
      LOGICAL               FAILED
      LOGICAL               VZERO
 
C
C     Statement function type declarations
C     
      DOUBLE PRECISION      NORMSQ
      INTEGER               VOX2ID 
      

C
C     Local parameters
C
      DOUBLE PRECISION      TOL
      PARAMETER           ( TOL   = 1.D-4 )

C
C     Maximum number of voxels we can accept for
C     one XDDA call.
C
      INTEGER               NVXLST
      PARAMETER           ( NVXLST = 50000 )
          
C
C     Maximum number of plates we work with
C     at any time.
C
      INTEGER               MAXPL
      PARAMETER           ( MAXPL  =  MAXPLT/125 )
          
C
C     Local Variables
C
      DOUBLE PRECISION      VTX2   ( 3 )
      DOUBLE PRECISION      UDIR   ( 3 )
      DOUBLE PRECISION      COORD  ( 3 )
      DOUBLE PRECISION      COSANG
      DOUBLE PRECISION      DIST
      DOUBLE PRECISION      EDGES  ( 3, 3 )
      DOUBLE PRECISION      MAG
      DOUBLE PRECISION      MAXT   ( 3 )
      DOUBLE PRECISION      NEAR
      DOUBLE PRECISION      NORML  ( 3 )
      DOUBLE PRECISION      OBSMAT ( 3, 3 )
      DOUBLE PRECISION      ORIGIN ( 3 )
      DOUBLE PRECISION      PERP   ( 3 )
      DOUBLE PRECISION      PLANE  ( 3 )
      DOUBLE PRECISION      PLTREC ( PLRCSZ )
      DOUBLE PRECISION      POINTS ( 3, 3 )
      DOUBLE PRECISION      PRVNRM ( 3 )
      DOUBLE PRECISION      PRVTOC ( 3 )
      DOUBLE PRECISION      SCALE
      DOUBLE PRECISION      TEMP
      DOUBLE PRECISION      TOCENT ( 3 )
      DOUBLE PRECISION      V1
      DOUBLE PRECISION      V2
      DOUBLE PRECISION      V3
      DOUBLE PRECISION      VOXFAR ( 3 )
      DOUBLE PRECISION      VOXORI ( 3 )
      DOUBLE PRECISION      VOXSIZ
      DOUBLE PRECISION      VTXBDS ( 2, 3 )


      INTEGER               CGREXT ( 3 )
      INTEGER               CGRPTR ( MAXCGR )
      INTEGER               CGSCAL
      INTEGER               CGSCL2
      INTEGER               CGXYZ  ( 3 )
      INTEGER               CVID
      INTEGER               DIM
      INTEGER               DIM1
      INTEGER               DIM2
      INTEGER               FINAL
      INTEGER               FX
      INTEGER               FY
      INTEGER               FZ
      INTEGER               GROUP
      INTEGER               GRPBEG
      INTEGER               GRPEND
      INTEGER               GRPSIZ
      INTEGER               I
      INTEGER               I1
      INTEGER               I2
      INTEGER               I3
      INTEGER               J
      INTEGER               K
      INTEGER               LEFT
      INTEGER               MIDDLE
      INTEGER               NCGR
      INTEGER               NGROUP
      INTEGER               NP
      INTEGER               NPLATE
      INTEGER               NV
      INTEGER               NVXOUT
      INTEGER               NVXTOT
      INTEGER               OFFSET
      INTEGER               ORDVEC ( MAXPL )
      INTEGER               PLANID
      INTEGER               PLATID ( MAXPL )
      INTEGER               PNTR
      INTEGER               PRVDSC ( DLADSZ )
      INTEGER               PRVHAN
      INTEGER               QUAD    ( 3 )
      INTEGER               RIGHT
      INTEGER               SOURCE ( MAXPL )
      INTEGER               START
      INTEGER               TEST
      INTEGER               TO
      INTEGER               TOTPLT
      INTEGER               VGREXT ( 3 )
      INTEGER               VI
      INTEGER               VIDS   ( 3 )
      INTEGER               VOXLST ( 3, NVXLST )
      INTEGER               VOXNPL
      INTEGER               VOXNPT
      INTEGER               VOXPTR
      INTEGER               VTXNPL
      INTEGER               VXLCG  ( 3, NVXLST )
      INTEGER               VXLOUT ( NVXLST )
      INTEGER               VXLSTR ( NVXLST )

      LOGICAL               HITS
      LOGICAL               HAVE
      LOGICAL               INHSPC
      LOGICAL               INSIDE
      LOGICAL               TOWARD
      LOGICAL               TOOFAR
      LOGICAL               EXTRA
      LOGICAL               NEWSEG

C
C     Saved variables
C
      SAVE                  CGREXT
      SAVE                  CGRPTR
      SAVE                  CGSCAL
      SAVE                  CGSCL2
      SAVE                  NCGR
      SAVE                  NP
      SAVE                  NV
      SAVE                  NVXTOT
      SAVE                  ORDVEC
      SAVE                  PLATID
      SAVE                  PRVHAN
      SAVE                  PRVDSC
      SAVE                  PRVNRM
      SAVE                  PRVTOC
      SAVE                  SOURCE
      SAVE                  VGREXT
      SAVE                  VOXLST
      SAVE                  VOXNPT
      SAVE                  VOXNPL
      SAVE                  VOXORI
      SAVE                  VOXSIZ
      SAVE                  VTXBDS
      SAVE                  VTXNPL
      SAVE                  VXLCG
      SAVE                  VXLOUT
      SAVE                  VXLSTR
      SAVE                  RIGHT
      SAVE                  LEFT
      SAVE                  MIDDLE

C
C     Initial values
C
      DATA                  RIGHT   / 0 /
      DATA                  LEFT    / 1 /
      DATA                  MIDDLE  / 2 /
      DATA                  PRVHAN  / 0 /
      DATA                  PRVDSC  / DLADSZ * 0 /

C
C     The following intializations are performed to
C     satisfy compilers that may consider these 
C     variables "possibly uninitialized":
C
      DATA                  PRVNRM  / 3 * 0.D0   /
      DATA                  PRVTOC  / 3 * 0.D0   /

C
C     Statement functions
C
      NORMSQ ( V1, V2, V3 ) = V1*V1 + V2*V2 + V3*V3

      VOX2ID ( I1, I2, I3, DIM1, DIM2 ) =  I1 + DIM1*( I2 +  I3* DIM2 
     .                                                    - (1 + DIM2) )

C
C     Standard SPICE error handling.
C
      IF ( RETURN() ) THEN
         RETURN
      END IF
 
      CALL CHKIN ( 'DSKX02' )

C
C     Until we have better knowledge we assume there is no intersection
C
      PLID   =  0
      FOUND  = .FALSE.
      HAVE   = .FALSE.
      INSIDE = .TRUE.
      NEAR   = DPMAX()

 
C
C     Check whether the ray direction vector is the zero vector.
C
      IF ( VZERO( RAYDIR ) ) THEN

         CALL SETMSG ( 'Ray direction is the zero vector.' )
         CALL SIGERR ( 'SPICE(RAYISZEROVECTOR)'            )
         CALL CHKOUT ( 'DSKX02'                            )
         RETURN

      END IF
 
C
C     Obtain the unit vector of the ray from the observer.
C
      CALL VHAT ( RAYDIR, UDIR )

C
C     Decide whether we're looking at a new segment.
C
      NEWSEG = .TRUE.

      IF ( HANDLE .EQ. PRVHAN ) THEN
C
C        The input handle matches the previous handle.  Note that the
C        initial value of PRVHAN is 0, which is never a valid handle,
C        so on the first pass, this test will fail.
C
         IF (       ( DLADSC(IBSIDX) .EQ. PRVDSC(IBSIDX) )
     .        .AND. ( DLADSC(DBSIDX) .EQ. PRVDSC(DBSIDX) )
     .        .AND. ( DLADSC(CBSIDX) .EQ. PRVDSC(CBSIDX) )  ) THEN
C
C           All of the DLA segment base addresses match.
C
            NEWSEG = .FALSE.

         END IF

      END IF


      IF ( NEWSEG ) THEN
C
C        Retrieve the voxel grid origin in model
C        units and calculate the farthest extent of the
C        voxel grid in voxel space.
C
         CALL DSKP02 ( HANDLE, DLADSC, NV,     NP,     NVXTOT,  
     .                 VTXBDS, VOXSIZ, VOXORI, VGREXT,
     .                 CGSCAL, VTXNPL, VOXNPT, VOXNPL         )

        
         IF ( CGSCAL .LT. 1 ) THEN

            CALL SETMSG ( 'Coarse grid scale = #; should be >= 1.' )
            CALL ERRINT ( '#', CGSCAL                              )
            CALL SIGERR ( 'SPICE(VALUEOUTOFRANGE)'                 )
            CALL CHKOUT ( 'DSKX02'                                 )
            RETURN

         END IF

C
C        Get the coarse voxel grid dimensions and the coarse voxel
C        occupancy flags.
C
         DO I = 1, 3
            CGREXT(I) = VGREXT(I) / CGSCAL
         END DO

         CGSCL2 = CGSCAL * CGSCAL
      
         NCGR = NVXTOT / CGSCAL**3

         IF ( NCGR .GT. MAXCGR ) THEN

            CALL SETMSG ( 'Coarse grid size NCGR = #. Buffer size = #' )
            CALL ERRINT ( '#', NCGR                                    )
            CALL ERRINT ( '#', MAXCGR                                  )
            CALL SIGERR ( 'SPICE(GRIDTOOLARGE)'                        )
            CALL CHKOUT ( 'DSKX02'                                     )
            RETURN

         END IF

         CALL DSKI02 ( HANDLE, DLADSC, KWCGPT, 1, MAXCGR, DIM, CGRPTR )

         PRVHAN = HANDLE

         CALL MOVEI ( DLADSC, DLADSZ, PRVDSC )

      END IF


C     The observer position in voxel space.
C
      CALL TOGRID ( VERTEX, VOXORI, VOXSIZ, ORIGIN )

C
C     The grid origin in voxel space lies at (0,0,0). Calculate
C     the farthest extent of the voxel grid in voxel space, i.e.
C     cast the grid dimension to double precision.
C
      VOXFAR(1) =  DBLE( VGREXT(1) )
      VOXFAR(2) =  DBLE( VGREXT(2) )
      VOXFAR(3) =  DBLE( VGREXT(3) )

C
C     Determine the intersection between the RAYDIR vector
C     and the voxel grid using the Digital Differential 
C     Analyzer Method.
C
C     Find the candidate plane.
C
      DO I = 1, 3

         IF (  ORIGIN(I) .LT. 0.D0 ) THEN

            QUAD(I) = LEFT
            PLANE(I)= 0.D0
            INSIDE  = .FALSE.

         ELSE IF  ( ORIGIN(I) .GT. VOXFAR(I) ) THEN

            QUAD(I) = RIGHT
            PLANE(I)= VOXFAR(I)
            INSIDE  = .FALSE.

         ELSE

            QUAD(I) = MIDDLE

         END IF

      END DO 

C
C     Check whether VERTEX exists within the voxel grid.
C
      IF ( .NOT. INSIDE ) THEN

C
C        Calculate T distances to candidate planes.
C
         DO I = 1, 3

            IF (  QUAD(I) .NE. MIDDLE .AND. RAYDIR(I) .NE. 0 ) THEN

               MAXT(I) = (PLANE(I) - ORIGIN(I) )/RAYDIR(I)

            ELSE

               MAXT(I) = -1.D0

            END IF

         END DO

C
C        Determine the largest of the MAXT for final choice of
C        intersection.
C
         PLANID = 1

         DO I = 2, 3

            IF ( MAXT(PLANID) .LT. MAXT(I) ) THEN
               PLANID = I
            END IF

         END DO

C
C        Confirm the final candidate exists inside the grid.
C
         IF( MAXT(PLANID) .LT. 0.D0 ) THEN

            FOUND = .FALSE.

            CALL CHKOUT ( 'DSKX02' )
            RETURN

         END IF

         DO I = 1, 3

            IF ( PLANID .NE. I ) THEN
C
C              Calculate the expected intersection coordinate
C              with the voxel grid over each degree-of-freedom.
C
               COORD(I) = ORIGIN(I) + MAXT(PLANID) * RAYDIR(I)
               
C
C              If the calculated COORD is outside the box,
C              a true intersection does not exist. Return
C              FALSE.
C
               COORD(I) = TOUCHD(COORD(I))

               
               IF(      ( COORD(I) .LT. -TOL          )
     .             .OR. ( COORD(I) .GT. VOXFAR(I)+TOL ) ) THEN

                  FOUND = .FALSE.

                  CALL CHKOUT ( 'DSKX02' )
                  RETURN

               END IF

            ELSE

               COORD(I) = PLANE(I)

            END IF

         END DO

C
C        Rather than using the original observer's location, we use
C        a location derived from COORD, which is the intercept of 
C        the ray and the surface of the voxel grid.  We start with
C        COORD, convert this location to the model coordinate system,
C        and back outward a bit to make sure we obtain a location
C        outside the grid (we don't want to miss any plates that
C        might be located right on the grid's surface).
C
         CALL VLCOM3 (  1.D0, VOXORI, VOXSIZ, COORD, 
     .                 -1.D0, UDIR, VTX2           )

      ELSE

C
C        Observer point exists inside voxel grid, no
C        calculations needed.
C
         COORD(1) = ORIGIN(1)
         COORD(2) = ORIGIN(2)
         COORD(3) = ORIGIN(3)

         CALL VEQU ( VERTEX, VTX2 )

      END IF

C
C     A tiny step down the boresight to overcome round-off problems.
C     Such problems can occur if the value for the intersection point
C     exceeds the box bounds due to round-off.
C
      MAG      = VNORM( RAYDIR )
      COORD(1) = COORD(1) + ( RAYDIR(1)* TOL )/MAG
      COORD(2) = COORD(2) + ( RAYDIR(2)* TOL )/MAG
      COORD(3) = COORD(3) + ( RAYDIR(3)* TOL )/MAG
                  
C
C     Need an error check on return size of VOXLST.
C
      CALL XDDA ( COORD, UDIR, VGREXT, NVXLST, NVXOUT, VOXLST )

C
C     Use the coarse voxel grid to compress the voxel list.  We
C     remove all voxels belonging to empty coarse voxels.
C
      TO = 0
  
      DO I = 1, NVXOUT
C
C        Find the coordinates, then the ID, of the coarse voxel
C        containing this voxel.
C         
         DO J = 1, 3
            CGXYZ(J) = ( ( VOXLST(J,I) - 1 ) / CGSCAL ) + 1
         END DO


         CVID = VOX2ID ( CGXYZ(1),  CGXYZ(2), CGXYZ(3), 
     .                   CGREXT(1), CGREXT(2)           )
         
         IF ( CGRPTR(CVID) .GT. 0 ) THEN
C
C           This coarse voxel is non-empty; add the index of the
C           current voxel to the output list.  Save the coordinates of
C           the parent coarse voxel as well.
C           
            TO         = TO + 1
            VXLOUT(TO) = I

            DO J = 1, 3

               VXLCG (J,TO) = CGXYZ (J)

            END DO

C
C           Save the coarse voxel start pointer as well.
C
            VXLSTR(TO) = CGRPTR(CVID)

         END IF

      END DO

C
C     Update NVXOUT to be the number of voxels in the compressed list.
C
      NVXOUT = TO
          
C
C     Break up the list of voxels into groups; process each 
C     group in turn until we find an intersection or run out
C     of voxels.
C
      GRPSIZ =  MAX ( 1,  (NVXOUT+1)/2)

      NGROUP =  (NVXOUT-1)/GRPSIZ + 1
          
      GROUP  =  1

      DO WHILE (  ( GROUP .LE. NGROUP )  .AND.  ( .NOT. HAVE )  )


         PNTR   = 1

         GRPBEG = (GROUP-1)*GRPSIZ + 1

         GRPEND = MIN ( GRPBEG + GRPSIZ - 1,  NVXOUT )


         DO VI = GRPBEG, GRPEND
C
C           Look up the plate list pointer for this voxel.
C
C
C           We begin by finding the offset of the voxel from
C           the base of its parent coarse voxel.
C
            J      =  VXLOUT(VI)

            FX     =  VOXLST(1,J)  -  CGSCAL * ( VXLCG(1,VI) - 1 )
            FY     =  VOXLST(2,J)  -  CGSCAL * ( VXLCG(2,VI) - 1 )
            FZ     =  VOXLST(3,J)  -  CGSCAL * ( VXLCG(3,VI) - 1 )

            OFFSET = FX + CGSCAL*(FY-1) + CGSCL2*(FZ-1)

C
C           Now compute the index of voxel-plate list pointer in
C           the pointer array, and look up the pointer.
C
            J = VXLSTR(VI) + OFFSET - 1

            CALL DSKI02 ( HANDLE, DLADSC, KWVXPT, J, 1, DIM, VOXPTR )


            IF ( VOXPTR .EQ. -1 ) THEN
               
               NPLATE = 0

            ELSE
C
C              Get the plate count for this voxel.
C
               CALL DSKI02 ( HANDLE, DLADSC, KWVXPL, VOXPTR, 
     .                       1,      DIM,    NPLATE          )
            END IF



            IF ( NPLATE .GT. 0 ) THEN

               IF ( NPLATE .LE. MAXPL ) THEN
C
C                 Get the plate list for this voxel.
C
                  CALL DSKI02 ( HANDLE,   DLADSC, KWVXPL,  
     .                          VOXPTR+1, NPLATE, DIM,    PLATID(PNTR) )
               ELSE

                  CALL SETMSG ( 'NPLATE = #. Buffer size = #'   )
                  CALL ERRINT ( '#', NPLATE                     )
                  CALL ERRINT ( '#', MAXPL                      )
                  CALL SIGERR ( 'SPICE(ARRAYTOOSMALL)'          )
                  CALL CHKOUT ( 'DSKX02'                        )
                  RETURN

               END IF

C
C              Fill in the corresponding elements of the parallel
C              "source" array with the current voxel loop index.
C              XDDA lists these voxels in the order the ray hits
C              them, so the lowest indexed voxels are hit first.
C
               CALL FILLI ( VI, NPLATE, SOURCE(PNTR) )

            END IF

C
C           NPLATE returns zero or greater.
C              
            PNTR = PNTR + NPLATE

         END DO
C
C        We've collected all plate info for the current voxel group.
C
         TOTPLT = PNTR-1

C
C        We want to sort the plate array by ID and remove
C        duplicates. However, we want to keep the plates ordered
C        according to the sequence in which their containing 
C        voxels were hit by the ray.  So we find the order vector for
C        the plate array, then use this vector to mark duplicates.  
C        We don't re-arrange the plate list itself since the plates are 
C        already listed in the order in which the ray hits their
C        containing voxels.
    
         CALL ORDERI ( PLATID, TOTPLT, ORDVEC )
C
C        Negate the plate ID of every duplicate we find.
C
         DO I = 2, TOTPLT
 
            IF (              PLATID( ORDVEC(I  ) )  
     .           .EQ.  ABS (  PLATID( ORDVEC(I-1) )  )    ) THEN
C
C              The plate having index ORDVEC(I) is a duplicate.
C
               PLATID( ORDVEC(I) ) = -PLATID( ORDVEC(I) )
                           
             END IF
 
         END DO
 
C
C        If something went wrong up above there is no point in
C        going on from here.
C 
         IF ( FAILED() ) THEN
            CALL CHKOUT ( 'DSKX02' )
            RETURN
         END IF
 
C
C        We are going to need to subtract the location of the observer
C        from vertices of a plate.  To speed things up a tiny bit,
C        we'll make 3 copies of the observer's location so that we make
C        a single subroutine call to handle the 3 subtractions.
C                                        
         CALL VEQU ( VTX2, OBSMAT(1,1) )
         CALL VEQU ( VTX2, OBSMAT(1,2) )
         CALL VEQU ( VTX2, OBSMAT(1,3) )
 
C
C        Now examine each plate in the PLATID list for this voxel group.
C        PNTR has the value of the index available for data in
C        PLATID, so the final location of data is at index PNTR - 1.
C
         EXTRA  = .FALSE.
         FINAL  = 0

         I      = 1

         DO WHILE (  ( I .LE. TOTPLT ) .AND. ( .NOT. EXTRA )  )
C
C           Retrieve the current plate ID.
C
            J = PLATID(I)

            IF ( J .GT. 0 ) THEN
C
C              This is not a duplicate plate; consider it.
C           
               IF ( HAVE ) THEN
C
C                 We already have a hit. See whether this plate resides
C                 in the voxel in which the last hit was found, or in a
C                 later voxel.
C
                  IF ( SOURCE(I) .NE. FINAL ) THEN
C
C                    This is a "late plate": it occurs in a voxel
C                    later than that in which the first hit was
C                    found.
C
                     EXTRA = .TRUE.

                  END IF

               END IF

               TOWARD = .FALSE.
               INHSPC = .FALSE.
               TOOFAR = .FALSE.


               IF ( .NOT. EXTRA ) THEN
C           
C                 Get the J'th plate record and plate normal.
C
                  START = PLRCSZ*(J-1) + 1

                  CALL DSKD02 ( HANDLE, DLADSC, KWPLRC, 
     .                          START,  PLRCSZ, DIM,   PLTREC )

                  CALL VEQU ( PLTREC(IXNRML), NORML )

C
C                 Throw out plates whose normals point away from the
C                 ray.
C
                  TOWARD =  VDOT ( UDIR, NORML ) .LE. 0.D0

C
C                 Only if this plate is visible do we do anything
C                 else with this plate.
C
                  IF ( TOWARD ) THEN
C
C                    Compute the position of the observer relative
C                    to the center of the plate. Optimize for speed.
C
C                    Get the J'th plate center and the vector from
C                    the observer to this center TOCENT.
C
                     CALL VSUB  ( PLTREC(IXCTR), VTX2, TOCENT )

C
C                    Find the distance of the plate center from the
C                    ray.  If this distance exceeds the length of
C                    the longest side of the plate, there can be no
C                    intersection.
C
                     CALL VPERP ( TOCENT, UDIR, PERP   )
               
                     TOOFAR =     NORMSQ( PERP(1), PERP(2), PERP(3) )
     .                       .GT. PLTREC( IXLSID )**2.D0


                     IF ( .NOT. TOOFAR ) THEN
C
C                       See whether the observer is in the "outer"
C                       half-space defined by the plate containing the
C                       plate and the plate's outward normal.
C
C                       This test is useful only if the observer is
C                       inside the voxel grid; otherwise, with rare
C                       exceptions, any plate rejected by this test
C                       would already have been rejected by the normal
C                       vector direction test.
C                                  
                        IF ( INSIDE ) THEN
                           INHSPC = VDOT ( TOCENT, NORML ) .LT. 0.D0
                        ELSE
                           INHSPC = .TRUE.
                        END IF

                     END IF

                  END IF

               END IF


               IF (        TOWARD 
     .              .AND.  INHSPC 
     .              .AND.  .NOT. TOOFAR 
     .              .AND.  .NOT. EXTRA  ) THEN
C                 
C                 The current plate qualifies for testing using INSANG.
C
C                 Retrieve the model coordinates of the J'th plate's
C                 three vertices. Compute the edges of the tetrahedral
C                 angle with the observer as the apex and the vertices
C                 as members of the edge rays. Finally see if the
C                 boresight lies inside the tetrahedron.
C
                  START = 3*(J-1) + 1
      
                  CALL DSKI02 ( HANDLE, DLADSC, KWPLAT, 
     .                          START,  3,      DIM,    VIDS )

                  DO K = 1, 3

                     START = 3*( VIDS(K) - 1 ) + 1

                     CALL DSKD02 ( HANDLE, DLADSC, KWVERT,  
     .                             START,  3,      DIM,    POINTS(1,K) )

                  END DO


                  CALL VSUBG  ( POINTS,     OBSMAT, 9,  EDGES )

                  CALL INSANG ( UDIR,
     .                          EDGES(1,1), EDGES(1,2), EDGES(1,3),
     .                          HITS,       SCALE )
 
C
C                 If the boresight intersects this plate, see if we
C                 already had an intersection.  If so see if this
C                 plate is closer than the last intersection we had.
C
                  IF ( HITS .AND. HAVE ) THEN 
C
C                    If this intersection point is closer than
C                    the last one, this plate must be in front.
C
                     COSANG = VDOT ( UDIR, PRVNRM )
                     DIST   = VNORM( PRVTOC )

                     IF ( SCALE .LT. NEAR ) THEN
 
                        NEAR   = SCALE
                        COSANG = VDOT ( UDIR, NORML )
                        PLID   = J

                     ELSE IF ( SCALE .EQ. NEAR ) THEN
C
C                       In this case we must have an edge point. We
C                       pick the plate whose normal direction
C                       is closest to anti-parallel to the boresight
C                       direction.

                        TEMP = VDOT ( UDIR, NORML )
  

                        IF ( TEMP .LT. COSANG ) THEN

                           COSANG = TEMP
                           DIST   = VNORM ( TOCENT ) 
                           PLID   = J
 
                        ELSE IF ( TEMP .EQ. COSANG ) THEN
C
C                          Hmmmm. We've got two parallel plates that
C                          share an edge or vertex.  Pick the one
C                          with the closest center.
C
                           TEMP = VNORM ( TOCENT )

                           IF ( TEMP .LT. DIST ) THEN

                              DIST = TEMP
                              PLID = J

                           ELSE IF ( TEMP .EQ. DIST ) THEN
C
C                             The centers of the plates are
C                             equidistant. That's the end of the
C                             geometric tests. We just take the one
C                             that had the lower plate ID number.
C
                              TEST = J

                              IF ( TEST .LT. PLID ) THEN
                                 PLID = TEST
                              END IF

                           END IF

                        END IF

                     END IF

                  ELSE IF ( HITS .AND. .NOT. HAVE ) THEN

                     HAVE   = .TRUE.
                     NEAR   = SCALE
                     PLID   = J
                     FINAL  = SOURCE(I)

C
C                    Save the plate normal and distance from
C                    VTX2 to the plate center; we may need these
C                    to pick the intercept plate if there are
C                    multiple hits.
C
                     CALL VEQU ( NORML,  PRVNRM )
                     CALL VEQU ( TOCENT, PRVTOC )

                  END IF
C
C                 We're done with processing the current hit.
C 
               END IF
C
C              We're done with processing the current qualifying plate.
C
            END IF
C
C           We're done with the current plate.              
C
C           Fetch the next plate for this voxel group.
C
            I = I + 1

         END DO
C
C        We're done with the current voxel group.
C
         GROUP = GROUP + 1

      END DO

C
C     We've done all of the tedious steps. If we have an intersection
C     put the information into FOUND.
C
      FOUND = HAVE
 
C
C     Compute intersection between the ray and plate, if the
C     intersection exists.
C
C         ______   ______        ______
C         XPT = VERTEX + NEAR*UDIR
C
      IF ( FOUND ) THEN
C
C             Use the new observer location as the origin of the
C        ray; construct the intercept point on the plate.
C
         CALL VLCOM ( 1.D0, VTX2, NEAR, UDIR, XPT )              

      END IF
C
C     That's all folks
C
      CALL CHKOUT ( 'DSKX02' )
      RETURN
      END
 
