C$Procedure MKSPIN ( Make spatial index of plates )
 
      SUBROUTINE MKSPIN ( NP,     VERT,
     .                    VOXSCL, CGSCAL, NVOX,   
     .                    VOXSIZ, VOXORI,
     .                    NVXTOT, VXPTR,  
     .                    NVXLST, VXLIST,
     .                    EXTENT, CGRPTR )
      IMPLICIT NONE

C$ Abstract
C
C     Index each plate with respect to a voxel grid.
C
C$ Disclaimer
C
C     THIS SOFTWARE AND ANY RELATED MATERIALS WERE CREATED BY THE
C     CALIFORNIA INSTITUTE OF TECHNOLOGY (CALTECH) UNDER A U.S.
C     GOVERNMENT CONTRACT WITH THE NATIONAL AERONAUTICS AND SPACE
C     ADMINISTRATION (NASA). THE SOFTWARE IS TECHNOLOGY AND SOFTWARE
C     PUBLICLY AVAILABLE UNDER U.S. EXPORT LAWS AND IS PROVIDED "AS-IS"
C     TO THE RECIPIENT WITHOUT WARRANTY OF ANY KIND, INCLUDING ANY
C     WARRANTIES OF PERFORMANCE OR MERCHANTABILITY OR FITNESS FOR A
C     PARTICULAR USE OR PURPOSE (AS SET FORTH IN UNITED STATES UCC
C     SECTIONS 2312-2313) OR FOR ANY PURPOSE WHATSOEVER, FOR THE
C     SOFTWARE AND RELATED MATERIALS, HOWEVER USED.
C
C     IN NO EVENT SHALL CALTECH, ITS JET PROPULSION LABORATORY, OR NASA
C     BE LIABLE FOR ANY DAMAGES AND/OR COSTS, INCLUDING, BUT NOT
C     LIMITED TO, INCIDENTAL OR CONSEQUENTIAL DAMAGES OF ANY KIND,
C     INCLUDING ECONOMIC DAMAGE OR INJURY TO PROPERTY AND LOST PROFITS,
C     REGARDLESS OF WHETHER CALTECH, JPL, OR NASA BE ADVISED, HAVE
C     REASON TO KNOW, OR, IN FACT, SHALL KNOW OF THE POSSIBILITY.
C
C     RECIPIENT BEARS ALL RISK RELATING TO QUALITY AND PERFORMANCE OF
C     THE SOFTWARE AND ANY RELATED MATERIALS, AND AGREES TO INDEMNIFY
C     CALTECH AND NASA FOR ALL THIRD-PARTY CLAIMS RESULTING FROM THE
C     ACTIONS OF RECIPIENT IN THE USE OF THE SOFTWARE.
C
C$ Required_Reading
C
C     None.
C
C$ Keywords
C
C     plate voxel index
C
C$ Declarations
  
      INCLUDE                'pltmax.inc'
      
      INTEGER               NP
      DOUBLE PRECISION      VERT     ( 3, 3, * )
      DOUBLE PRECISION      VOXSCL
      INTEGER               CGSCAL
      INTEGER               NVOX     ( 3 )
      DOUBLE PRECISION      VOXSIZ
      DOUBLE PRECISION      VOXORI   ( 3 )
      INTEGER               NVXTOT
      INTEGER               VXPTR    ( * )
      INTEGER               NVXLST
      INTEGER               VXLIST   ( * )
      DOUBLE PRECISION      EXTENT   ( 6 )
      INTEGER               CGRPTR   ( * )
 
C$ Brief_I/O
C
C     VARIABLE  I/O  DESCRIPTION
C     --------  ---  --------------------------------------------------
C     NP         I   Number of plates.
C     VERT       I   Array of all vertex coordinates.
C     VOXSCL     I   Voxel scaling factor.
C     CGSCAL     I   Coarse voxel grid scaling factor.
C     NVOX       O   Dimensions of voxel grid.
C     VOXSIZ     O   Size of a voxel in model units.
C     VOXORI     O   Origin of voxel grid in model units.
C     NVXTOT     O   Total number of voxels in grid.
C     VXPTR      O   Pointer index to the voxel to plates list.
C     NVXLST     O   Length of voxel to plates list.
C     VXLIST     O   Voxel to plates list.
C
C$ Detailed_Input
C
C     NP         Total number of plates in model.
C
C     VERT       Array containing XYZ coordinates for all vertices
C                in the model reference frame.
C
C     VOXSCL     The voxel size is determined by the
C                average extent of each plate in XYZ.  VOXSCL is a
C                scaling factor that may be used to adjust
C                final voxel size.
C
C     CGSCAL     The coarse voxel grid has voxel edge length
C                larger by factor CGSCAL than the fine voxel
C                edge length.
C
C$ Detailed_Output
C
C     NVOX       Dimension of the voxel grid in voxel units.
C
C     VOXSIZ     Size of each voxel in model units.
C
C     VOXORI     Origin of voxel grid in model units.
C
C     NVXTOT     Total number of voxels in grid.
C
C     VXPTR      Pointer index to the voxel to plates list.
C                The Nth element of this array contains a
C                pointer to the Mth element of VXLIST, or
C                -1.  If VXPTR(N) is -1, the Nth voxel is empty.
C
C     NVXLST     The total number of voxels in the grid,
C                VVXLST = NVOX(1) * NVOX(2) * NVOX(3).
C
C     VXLIST     The voxel to plates list.  For the Nth voxel,
C                VXPTR(N) points to VXLIST(M).  VXLIST(M) contains
C                the number plates that might be found in voxel N,
C                and the plates IDs are stored in the following
C                elements.
C
C$ Parameters
C
C     None.
C
C$ Exceptions
C
C     NOT Error free.
C
C$ Files
C
C     None.
C
C$ Particulars
C
C     Later.
C
C$ Examples
C
C     Later.
C
C$ Restrictions
C
C     None.
C
C$ Literature_References
C
C     None.
C
C$ Author_and_Institution
C
C     J.A. Bytof      (JPL)
C
C$ Version
C
C-    SPICELIB Version 1.1.1, 08-OCT-2009 (NJB)
C
C        Re-ordered header sections.
C
C-    SPICELIB Version 1.1.0, 12-SEP-2004 (EDW)
C
C       Improve/expand comments and descriptions.
C
C-    SPICELIB Version 1.0.0, 03-FEB-1999 (JAB)
C
C-&
 
C$ Index_Entries
C
C     spatial index plates voxels
C
C-&


C
C     SPICELIB functions
C
      DOUBLE PRECISION      DPMAX
      DOUBLE PRECISION      DPMIN

      INTEGER               VOX2ID

      LOGICAL               RETURN

C
C     Local parameters
C

C
C     Fraction of voxel edge length used a tolerance for plate
C     inclusion in voxels:
C
      DOUBLE PRECISION      TOL
      PARAMETER           ( TOL  = 1.D-3 )

C
C     Local Variables
C
      DOUBLE PRECISION      AVEXT
      DOUBLE PRECISION      BXMAX
      DOUBLE PRECISION      BXMIN
      DOUBLE PRECISION      BYMAX
      DOUBLE PRECISION      BYMIN
      DOUBLE PRECISION      BZMAX
      DOUBLE PRECISION      BZMIN
      DOUBLE PRECISION      CVXSIZ
      DOUBLE PRECISION      MDLTOL
      DOUBLE PRECISION      VMOD   ( 3 )
      DOUBLE PRECISION      VOXPAD
      DOUBLE PRECISION      XMAX
      DOUBLE PRECISION      XMIN
      DOUBLE PRECISION      XP     ( 3 )
      DOUBLE PRECISION      XVMAX
      DOUBLE PRECISION      XVMIN
      DOUBLE PRECISION      YMAX
      DOUBLE PRECISION      YMIN
      DOUBLE PRECISION      YP     ( 3 )
      DOUBLE PRECISION      YVMAX
      DOUBLE PRECISION      YVMIN
      DOUBLE PRECISION      ZMAX
      DOUBLE PRECISION      ZMIN
      DOUBLE PRECISION      ZP     ( 3 )
      DOUBLE PRECISION      ZVMAX
      DOUBLE PRECISION      ZVMIN

      INTEGER               CELLS  ( 2, MAXNPV ) 
      INTEGER               CGRDIM ( 3 )
      INTEGER               CVID
      INTEGER               GXMAX
      INTEGER               GXMIN
      INTEGER               GYMAX
      INTEGER               GYMIN
      INTEGER               GZMAX
      INTEGER               GZMIN
      INTEGER               I
      INTEGER               IX
      INTEGER               IY
      INTEGER               IZ
      INTEGER               J
      INTEGER               K
      INTEGER               L
      INTEGER               M
      INTEGER               N
      INTEGER               N1VOX
      INTEGER               NCELL
      INTEGER               NCGFLG
      INTEGER               NEMPTY
      INTEGER               NX
      INTEGER               NY
      INTEGER               NZ
      INTEGER               PNTRS  ( MAXNPV )
      INTEGER               PTR
      INTEGER               VID
      INTEGER               VIXYZ  ( 3 )
      INTEGER               VCOORD ( 3 )
      
      LOGICAL               FOUND
      LOGICAL               INBOX

C
C     Saved variables
C
C
C     Required for f2c use on Linux, all local variables
C     to static.
C 
      SAVE


C
C     Standard SPICE error handling.
C
      IF ( RETURN () ) THEN
         RETURN
      ELSE
         CALL CHKIN ( 'MKSPIN' )
      END IF

C
C     Get the average extents of all plates and the
C     overall model extent.
C 
      AVEXT = 0.D0
 
      XMIN =  DPMAX()
      XMAX =  DPMIN()
      YMIN =  DPMAX()
      YMAX =  DPMIN()
      ZMIN =  DPMAX()
      ZMAX =  DPMIN()
 
      DO I = 1, NP
 
         BXMIN = DPMAX()
         BXMAX = DPMIN()
         BYMIN = DPMAX()
         BYMAX = DPMIN()
         BZMIN = DPMAX()
         BZMAX = DPMIN()
 
         XP(1) = VERT(1,1,I)
         XP(2) = VERT(1,2,I)
         XP(3) = VERT(1,3,I)
 
         YP(1) = VERT(2,1,I)
         YP(2) = VERT(2,2,I)
         YP(3) = VERT(2,3,I)
 
         ZP(1) = VERT(3,1,I)
         ZP(2) = VERT(3,2,I)
         ZP(3) = VERT(3,3,I)
 
         DO J = 1, 3

C
C           Determine plate extents.
C
            BXMIN = DMIN1( BXMIN, XP(J) )
            BXMAX = DMAX1( BXMAX, XP(J) )
 
            BYMIN = DMIN1( BYMIN, YP(J) )
            BYMAX = DMAX1( BYMAX, YP(J) )
 
            BZMIN = DMIN1( BZMIN, ZP(J) )
            BZMAX = DMAX1( BZMAX, ZP(J) )
 
C
C           Determine model extent.
C
            XMIN = DMIN1 ( XMIN, BXMIN )
            XMAX = DMAX1 ( XMAX, BXMAX )
 
            YMIN = DMIN1 ( YMIN, BYMIN )
            YMAX = DMAX1 ( YMAX, BYMAX )
 
            ZMIN = DMIN1 ( ZMIN, BZMIN )
            ZMAX = DMAX1 ( ZMAX, BZMAX )
 
         END DO
 
         EXTENT(1) = XMIN
         EXTENT(2) = XMAX
         EXTENT(3) = YMIN
         EXTENT(4) = YMAX
         EXTENT(5) = ZMIN
         EXTENT(6) = ZMAX

         
C
C        Calculate the cumulative extent of all plates for
C        each degree of freedom.
C
         AVEXT = AVEXT + DABS(BXMAX-BXMIN) +
     .                   DABS(BYMAX-BYMIN) +
     .                   DABS(BZMAX-BZMIN)
 
      END DO

C
C     Calculate the average extent of all plates for
C     and the voxel size, i.e the length of one side
C     of a voxel cube.
C
      AVEXT  = AVEXT / DBLE(3*NP)
      VOXSIZ = VOXSCL * AVEXT
      MDLTOL = VOXSIZ * TOL

C      WRITE(*,*) 'Average plate extent:  ', AVEXT
C      WRITE(*,*) 'Fine voxel size:       ', VOXSIZ
C          WRITE(*,*) 'Tolerance in km:       ', MDLTOL

      
C
C     Determine the size of the coarse voxels.
C
      CVXSIZ = VOXSIZ * CGSCAL

C      WRITE(*,*) 'Coarse voxel scale:    ', CGSCAL
C      WRITE(*,*) 'Coarse voxel size:     ', CVXSIZ


C
C     Determine the minima and maxima of the body centered
C     vertex coordinates expressed in coarse voxel units. Scale the 
C     vertices coord values by CVXSIZ, this scales the 
C     axis in the voxel model space producing cubic voxels
C     with length 1 along each edge in voxel space,
C     VOXSIZ along an edge in model space.
C
      XVMIN = XMIN / CVXSIZ
      YVMIN = YMIN / CVXSIZ 
      ZVMIN = ZMIN / CVXSIZ
      XVMAX = XMAX / CVXSIZ
      YVMAX = YMAX / CVXSIZ
      ZVMAX = ZMAX / CVXSIZ
 
C
C     Extend the coarse voxel grid by 1 voxel length along each 
C     degree of freedom.
C
      XVMIN = VOXPAD ( -1, XVMIN )
      YVMIN = VOXPAD ( -1, YVMIN )
      ZVMIN = VOXPAD ( -1, ZVMIN )
      XVMAX = VOXPAD (  1, XVMAX )
      YVMAX = VOXPAD (  1, YVMAX )
      ZVMAX = VOXPAD (  1, ZVMAX )
 
 
C
C     Calculate the coarse voxel grid origin in model units.
C
      VOXORI(1) = XVMIN * CVXSIZ
      VOXORI(2) = YVMIN * CVXSIZ
      VOXORI(3) = ZVMIN * CVXSIZ
 
C
C     Calculate the dimension of the voxel grid in
C     units of (regular) voxels.
C
      NX = NINT( XVMAX - XVMIN ) * CGSCAL
      NY = NINT( YVMAX - YVMIN ) * CGSCAL
      NZ = NINT( ZVMAX - ZVMIN ) * CGSCAL
 
      NVOX(1) = NX
      NVOX(2) = NY
      NVOX(3) = NZ
 
      NVXTOT = NX*NY*NZ

C      WRITE(*,*) 'Voxel extents:     ', NVOX
C      WRITE(*,*) 'Total voxel count: ', NVXTOT
C
C     Enumerate all voxels that each plate might intersect.
C
      CALL INILNK ( MAXNPV, MAXNPV, NCELL, PNTRS, CELLS )

      N1VOX = 0

      DO I = 1, NP

C         WRITE (*,*) 'plate = ', I

         GXMIN =  MAXPLT + 1
         GXMAX =  -1
         GYMIN =  MAXPLT + 1
         GYMAX =  -1
         GZMIN =  MAXPLT + 1
         GZMAX =  -1

C
C        Loop over the plate's three vertices.
C
 
         DO J = 1, 3
C
C           Calculate the voxel coordinates for the voxels 
C           containing the body fixed vertex coordinates.
C
            CALL GETVOX ( VOXSIZ ,     VOXORI, NVOX, 
     .                    VERT(1,J,I), INBOX,  VCOORD )
 
            IF ( INBOX ) THEN
 
               GXMIN = MIN0( GXMIN, VCOORD(1) )
               GXMAX = MAX0( GXMAX, VCOORD(1) )
               GYMIN = MIN0( GYMIN, VCOORD(2) )
               GYMAX = MAX0( GYMAX, VCOORD(2) )
               GZMIN = MIN0( GZMIN, VCOORD(3) )
               GZMAX = MAX0( GZMAX, VCOORD(3) )
 
            ELSE

C
C              A plate vertex lies outside the voxel grid.
C              This should never signal. To do so indicates a
C              serious bug.
C
               CALL SETMSG ( 'BUG: Plate outside of voxel '//
     .                       'grid. plate ID = #.'          )
               CALL ERRINT ( '#', I                         )
               CALL SIGERR ( 'SPICE(PLATEOUTSIDEGRID)'      )
               CALL CHKOUT ( 'MKSPIN'                       )
               RETURN

            END IF
 
         END DO

C
C        Extend the vertex by TOL in each direction, if we
C        can do so without exiting the voxel grid.
C
         DO J = 1, 3

            DO K = -1, 1, 2

               DO L = -1, 1, 2

                  DO M = -1, 1, 2

                     VMOD(1) = VERT(1,J,I) + K*MDLTOL
                     VMOD(2) = VERT(2,J,I) + L*MDLTOL
                     VMOD(3) = VERT(3,J,I) + M*MDLTOL
  

                     CALL GETVOX ( VOXSIZ, VOXORI, NVOX, 
     .                             VMOD,   INBOX,  VCOORD )

                   
                     IF ( INBOX ) THEN

                        GXMIN = MIN0( GXMIN, VCOORD(1) )
                        GXMAX = MAX0( GXMAX, VCOORD(1) )
                        GYMIN = MIN0( GYMIN, VCOORD(2) )
                        GYMAX = MAX0( GYMAX, VCOORD(2) )
                        GZMIN = MIN0( GZMIN, VCOORD(3) )
                        GZMAX = MAX0( GZMAX, VCOORD(3) )

                     END IF

                  END DO

               END DO

            END DO

         END DO
                 

         IF (       ( GXMIN .EQ. GXMAX )
     .        .AND. ( GYMIN .EQ. GYMAX )
     .        .AND. ( GZMIN .EQ. GZMAX ) ) THEN

            N1VOX = N1VOX + 1

         END IF

C
C        Determine the other voxels, if any, that the plate
C        might intersect, based on the voxels that the vertices
C        of the plate lie in.
C
         DO IZ = GZMIN, GZMAX

             DO IY = GYMIN, GYMAX

                 DO IX = GXMIN, GXMAX

                    VIXYZ(1) = IX
                    VIXYZ(2) = IY
                    VIXYZ(3) = IZ
                    CALL ADDLNK ( VOX2ID(VIXYZ,NVOX), I, MAXNPV, 
     .                            PNTRS, NCELL, CELLS )
C                    WRITE (*,*) 'NCELL = ', NCELL
                 END DO

              END DO

          END DO

      END DO
          

C      WRITE(*,*) 'Number of plates in just one voxel:  ', N1VOX
C      WRITE(*,*) 'Total number of cells used: ', NCELL

C
C     Generate two linked lists mapping voxel ID to the plates enclosed
C     within that voxel (if any).
C
C     VXPTR : An array, indexed by voxel ID. For an array element, 
C             VXPTR(VOX_ID), greater than zero, the value identifies an 
C             index in VXLIST, the value of that VXLIST array elemement 
C             equaling the number of plates contained in the voxel 
C             specified by the ID. The condition VXPTR(VOX_ID) = -1 
C             indicates the voxel contains no plates.
C
C     VXLIST: An array, indexed by the positive entries in VXPTR. The 
C             element, N, identified by a VXPTR value describes the 
C             number of plates contained in the corresponding voxel. 
C
C                 N = VXLIST( VXPTR(VOX_ID) )
C
C             The N elements following VXLIST( VXPTR(VOX_ID) ),
C             contain the IDs of those plates within the voxel.
C
      CALL UNTNGL ( NVXTOT, PNTRS,  MAXNPV, CELLS, 
     .              VXPTR,  NVXLST, VXLIST         )

      NEMPTY = 0

      DO I = 1, NVXTOT
         IF ( VXPTR(I) .EQ. -1 ) THEN
            NEMPTY = NEMPTY + 1
         END IF
      END DO

C      WRITE (*,*) 'After UNTNGL call:'
C      WRITE (*,*) 'MAXNPV = ', MAXNPV
C      WRITE (*,*) 'NVXLST = ', NVXLST
C      WRITE (*,*) 'NP     = ', NP
C      WRITE (*,*) 'Number of empty voxels     = ', NEMPTY
C      WRITE (*,*) 'Fraction of empty voxels   = ', NEMPTY/DBLE(NVXTOT)
C      WRITE (*,*) 'Number of non-empty voxels = ', NVXTOT-NEMPTY
C      WRITE (*,*) 'Average number of links per non-empty voxel = ', 
C     .            DBLE(NVXLST)/DBLE(NVXTOT-NEMPTY) 
C      WRITE (*,*) 'Ratio of links to plates = ', DBLE(NVXLST)/NP

C      WRITE (*,*) 'Ratio of non-empty voxels to total voxel count = ', 
C     .            DBLE(NVXTOT-NEMPTY)/NVXTOT

      
      

      NCGFLG = NVXTOT / ( CGSCAL ** 3 )

C
C     Use the voxel-plate mapping to fill in the coarse grid flag
C     array.
C
      CGRDIM(1) = NX / CGSCAL
      CGRDIM(2) = NY / CGSCAL
      CGRDIM(3) = NZ / CGSCAL

C
C     Initialize the coarse grid flags to indicate all coarse
C     voxels are unoccupied.
C
      DO I = 1, NCGFLG
         CGRPTR(I) = 0
      END DO

      NEMPTY = NCGFLG
C
C     Assign a value to each coarse voxel flag.
C
      DO I = 1, CGRDIM(1)

         DO J = 1, CGRDIM(2)

            DO K = 1, CGRDIM(3)
C
C              Get the coarse voxel index for the current voxel.
C
               CALL IPACK3 ( I, J, K, VIXYZ )
               
               CVID = VOX2ID ( VIXYZ, CGRDIM )

C
C              Check each regular voxel in the current coarse
C              voxel until we find a non-empty one, or until
C              we've seen them all.
C
               FOUND = .FALSE.

               L     =  1

               DO WHILE (  ( L .LE. CGSCAL ) .AND. ( .NOT. FOUND )  )

                  M = 1

                  DO WHILE (  ( M .LE. CGSCAL ) .AND. ( .NOT. FOUND )  )

                     N = 1


                     DO WHILE (       ( N .LE.  CGSCAL ) 
     .                          .AND. (   .NOT. FOUND  )  )

C
C                       Compute the voxel grid coordinates of the
C                       current regular voxel.
C                         
                        IX = L + (I-1) * CGSCAL
                        IY = M + (J-1) * CGSCAL
                        IZ = N + (K-1) * CGSCAL
                      
C
C                       Get the voxel index for the current voxel.
C
                        CALL IPACK3 ( IX, IY, IZ, VIXYZ )
               
                        VID = VOX2ID ( VIXYZ, NVOX )

C
C                       Find out if the plate count for this voxel
C                       is positive.
C
                        PTR   = VXPTR  ( VID )

                        IF ( VXLIST(PTR) .GT. 0 ) THEN

                           FOUND        = .TRUE.
                           CGRPTR(CVID) = 1
                           
                           NEMPTY = NEMPTY - 1

                        ELSE
  
                           N = N + 1

                        END IF

                     END DO


                     M = M + 1

                  END DO

                  L = L + 1

               END DO

            END DO

         END DO

      END DO

C      WRITE (*,*) 'Coarse grid dimensions: ', CGRDIM
C      WRITE (*,*) 'Voxel  grid dimensions: ', NVOX
C      WRITE (*,*) 'Fraction of empty coarse voxels: ',
C     .            DBLE(NEMPTY) / DBLE(NCGFLG)


C
C     Checkout and return.
C
      CALL CHKOUT ( 'MKSPIN' )

      RETURN
 
      END
