C$Procedure PLCALC ( Plate Normal, Center, Area and Longest Side )
 
      SUBROUTINE PLCALC ( NP, MAXPLT, VERT, NORM, CENT, AREA, LSIDE )
      IMPLICIT NONE
 
C$ Abstract
C
C    Compute the unit normal vector of a triangular "plate" in 3-space.
C
C    Also return its area, extent and length of longest side.
C
C$ Disclaimer
C
C     THIS SOFTWARE AND ANY RELATED MATERIALS WERE CREATED BY THE
C     CALIFORNIA INSTITUTE OF TECHNOLOGY (CALTECH) UNDER A U.S.
C     GOVERNMENT CONTRACT WITH THE NATIONAL AERONAUTICS AND SPACE
C     ADMINISTRATION (NASA). THE SOFTWARE IS TECHNOLOGY AND SOFTWARE
C     PUBLICLY AVAILABLE UNDER U.S. EXPORT LAWS AND IS PROVIDED "AS-IS"
C     TO THE RECIPIENT WITHOUT WARRANTY OF ANY KIND, INCLUDING ANY
C     WARRANTIES OF PERFORMANCE OR MERCHANTABILITY OR FITNESS FOR A
C     PARTICULAR USE OR PURPOSE (AS SET FORTH IN UNITED STATES UCC
C     SECTIONS 2312-2313) OR FOR ANY PURPOSE WHATSOEVER, FOR THE
C     SOFTWARE AND RELATED MATERIALS, HOWEVER USED.
C
C     IN NO EVENT SHALL CALTECH, ITS JET PROPULSION LABORATORY, OR NASA
C     BE LIABLE FOR ANY DAMAGES AND/OR COSTS, INCLUDING, BUT NOT
C     LIMITED TO, INCIDENTAL OR CONSEQUENTIAL DAMAGES OF ANY KIND,
C     INCLUDING ECONOMIC DAMAGE OR INJURY TO PROPERTY AND LOST PROFITS,
C     REGARDLESS OF WHETHER CALTECH, JPL, OR NASA BE ADVISED, HAVE
C     REASON TO KNOW, OR, IN FACT, SHALL KNOW OF THE POSSIBILITY.
C
C     RECIPIENT BEARS ALL RISK RELATING TO QUALITY AND PERFORMANCE OF
C     THE SOFTWARE AND ANY RELATED MATERIALS, AND AGREES TO INDEMNIFY
C     CALTECH AND NASA FOR ALL THIRD-PARTY CLAIMS RESULTING FROM THE
C     ACTIONS OF RECIPIENT IN THE USE OF THE SOFTWARE.
C
C$ Required_Reading
C
C     None.
C
C$ Keywords
C
C     GEOMETRY
C
C$ Declarations

      INTEGER               NP
      INTEGER               MAXPLT
      DOUBLE PRECISION      VERT     ( 3, 3, * )
      DOUBLE PRECISION      NORM     ( 3, * )
      DOUBLE PRECISION      CENT     ( 3, * )
      DOUBLE PRECISION      AREA     ( * )
      DOUBLE PRECISION      LSIDE    ( * )

C$ Brief_I/O
C
C     VARIABLE  I/O  DESCRIPTION
C     --------  ---  --------------------------------------------------
C     NP         I   Number of plates.
C     MAXPLT     I   Max number of allowed plates; declared size of
C                    VERT, NORM, CENT, AREA, LSIDE.
C     VERT       I   X, Y and Z coordinates of each plate's vertex.
C     NORM       O   Unit normal vector of the plate.
C     CENT       O   Coordinates of the plate's center.
C     AREA       O   Plate surface area.
C     LSIDE      O   Length of each plate's longest side.
C
C
C$ Detailed_Input
C
C     TBD
C
C$ Detailed_Output
C
C     NORM       vector of unit normal vectors for the plates
C                at CENTER.
C
C     CENT       vector of the coordinates for the plates' center.
C
C     AREA       vector of the plate's surface area.
C
C     LSIDE      vector of the length for the plates' longest side.
C
C$ Parameters
C
C     None.
C
C$ Exceptions
C
C     1) The routine signals the error SPICE(ARRAYTOOSMALL)
C        if the number of plates (NP) to process exceeds
C        the size (MAXPLT) of input arrays.
C
C$ Files
C
C     None.
C
C$ Particulars
C
C     TBD
C
C$ Examples
C
C     TBD
C
C$ Restrictions
C
C     None.
C
C$ Literature_References
C
C     None.
C
C$ Author_and_Institution
C
C     J.A. Bytof      (JPL)
C
C$ Version
C
C-    SPICELIB Version 1.0.1, 08-OCT-2009 (NJB)
C
C        Re-ordered header sections.
C
C-    SPICELIB Version 1.0.0, 18-OCT-1998 (JAB)
C
C
C-&
 
C$ Index_Entries
C
C     Compute the normal, center, area and longest side of a plate
C
C-&
C
C     Local Variables
C
 
      INTEGER               I
 
      DOUBLE PRECISION      VNORM
      DOUBLE PRECISION      V1MAG
      DOUBLE PRECISION      V2MAG
      DOUBLE PRECISION      V3MAG
      DOUBLE PRECISION      VUPMAG
      DOUBLE PRECISION      V1        ( 3 )
      DOUBLE PRECISION      V2        ( 3 )
      DOUBLE PRECISION      V3        ( 3 )
      DOUBLE PRECISION      XP        ( 3 )
      DOUBLE PRECISION      YP        ( 3 )
      DOUBLE PRECISION      ZP        ( 3 )
      DOUBLE PRECISION      VUP       ( 3 )

      LOGICAL               RETURN
     

C
C     No need to proceed if in RETURN mode.
C
      IF ( RETURN() ) THEN
         RETURN
      END IF

      CALL CHKIN  ( 'PLCALC' )

C
C     Check array size versus loop extent.
C
      IF ( NP .GT. MAXPLT ) THEN

         CALL SETMSG ( 'Index larger than array. '//
     .                 'Index = #1. Array size = #2.')
         CALL ERRINT ( '#1', NP                      )
         CALL ERRINT ( '#2', MAXPLT                  )
         CALL SIGERR ( 'SPICE(ARRAYTOOSMALL)'        )
         CALL CHKOUT ( 'PLCALC'                      )
         RETURN

      END IF


C
C     Loop over all plates
C 
      DO I = 1, NP
 
         XP(1) = VERT(1,1,I)
         XP(2) = VERT(1,2,I)
         XP(3) = VERT(1,3,I)
 
         YP(1) = VERT(2,1,I)
         YP(2) = VERT(2,2,I)
         YP(3) = VERT(2,3,I)
 
         ZP(1) = VERT(3,1,I)
         ZP(2) = VERT(3,2,I)
         ZP(3) = VERT(3,3,I)
 
C
C        Form the vectors V1, V2 and V3 which point between
C        vertices 1-2, 2-3 and 3-1, respectively.
C
         V1(1) = XP(2) - XP(1)
         V1(2) = YP(2) - YP(1)
         V1(3) = ZP(2) - ZP(1)
 
         V2(1) = XP(3) - XP(2)
         V2(2) = YP(3) - YP(2)
         V2(3) = ZP(3) - ZP(2)
 
         V3(1) = XP(1) - XP(3)
         V3(2) = YP(1) - YP(3)
         V3(3) = ZP(1) - ZP(3)
 
C
C        Find the length of the longest side.
C
         V1MAG = VNORM( V1 )
         V2MAG = VNORM( V2 )
         V3MAG = VNORM( V3 )
 
         LSIDE(I) = DMAX1 ( V1MAG, V2MAG, V3MAG )
 
C
C        Determine the cross-product of vectors V1 and V2 -
C        this is the plate's normal vector.
C
         CALL VCRSS ( V1,  V2,        VUP    )
         CALL UNORM ( VUP, NORM(1,I), VUPMAG )

C
C        The area of the paralellogram defined by V1 and V2
C        equals the magnitude of the cross product, so 0.5 * VUP 
C         is the plate's area.
C
         AREA(I) = 0.5D0 * VUPMAG

C   
C        Average the values of each component over all vertices
C        to obtain the plate's centroid.
C
C        Note, add parameter for the 1/3 calculation.
C
         CENT(1,I) = ( XP(1) + XP(2) + XP(3) ) / 3.D0
         CENT(2,I) = ( YP(1) + YP(2) + YP(3) ) / 3.D0
         CENT(3,I) = ( ZP(1) + ZP(2) + ZP(3) ) / 3.D0
 
      END DO

      CALL CHKOUT ( 'PLCALC' )

      RETURN
      END
